/*
 * Copyright 2008,2009,2011,2012, Petr Machata
 *
 * This file is part of MicroSweeper.
 *
 * MicroSweeper is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * MicroSweeper is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MicroSweeper.  If not, see <http://www.gnu.org/licenses/>.
 */

public abstract class MineField {

    static private int lastid = 0;
    public final int id = lastid++;
    protected Field[] fields = null;
    protected final ParticipantSelector parSel;
    protected final BitSet128 flags;
    protected final BitSet128 uncovered;
    protected final BitSet128 mines;

    private boolean busted = false;
    private boolean won = false;

    protected MineField(ParticipantSelector parSel)
	throws IllegalArgumentException
    {
	this.parSel = parSel;
	this.flags = new BitSet128();
	this.uncovered = new BitSet128();
	this.mines = new BitSet128();
    }

    protected Field[] buildFieldArray(int numberOfFields) {
	Field[] fields = new Field[numberOfFields];
	for (byte i = 0; i < numberOfFields; ++i)
	    fields[i] = new NormalField(this, i);
	return fields;
    }

    protected interface TotalsBuilder {
	public Field buildColumnField(byte idx);
	public Field buildRowField(byte idx);
	public byte rowFieldIdx(byte row);
    }

    protected void addTotals(int w, int h,
			     boolean sums, boolean total,
			     TotalsBuilder totBld) {
	byte totalIndex = 0;
	if (sums) {
	    for (byte idx = 1; idx < w; ++idx)
		this.fields[idx] = totBld.buildColumnField(idx);

	    for (byte i = 1; i < h; ++i) {
		byte idx = totBld.rowFieldIdx(i);
		this.fields[idx] = totBld.buildRowField(idx);
	    }

	} else {
	    totalIndex = (byte)(this.fields.length - 1);
	}

	if (total)
	    this.fields[totalIndex] = new TotalCountField(this, totalIndex);
	else
	    this.fields[totalIndex] = new EmptyField(this, totalIndex);
    }

    public MineField(MineField copy) {
	int s = copy.fields.length;
	fields = new Field[s];
	for (int i = 0; i < s; ++i)
	    fields[i] = copy.fields[i].clone(this);
	this.parSel = copy.parSel;
	this.flags = new BitSet128(copy.flags);
	this.uncovered = new BitSet128(copy.uncovered);
	this.mines = new BitSet128(copy.mines);
    }

    public boolean flagged(Field field) {
	return flags.containsKey(field.fieldIdx);
    }

    public void toggleFlag(Field field) {
	if (!flags.remove(field.fieldIdx))
	    flags.add(field.fieldIdx);
    }

    final static int UNCOVER_BUSTED = -1;
    final static int UNCOVER_OLD = 0;
    final static int UNCOVER_NEW = 1;

    public boolean uncovered(Field field) {
	return uncovered.containsKey(field.fieldIdx);
    }

    public int uncover(Field field) {
	if (uncovered(field))
	    return UNCOVER_OLD;

	uncovered.add(field.fieldIdx);
	if (field.hasMine())
	    return UNCOVER_BUSTED;

	return UNCOVER_NEW;
    }

    public boolean hasMine(Field field) {
	return mines.containsKey(field.fieldIdx);
    }

    public void setMine(Field field, boolean whether) {
	if (whether)
	    mines.add(field.fieldIdx);
	else
	    mines.remove(field.fieldIdx);
    }

    abstract public MineField clone();

    // - - - - - - - - - - - - - - - - -
    // Coordinate handling
    // - - - - - - - - - - - - - - - - -

    // Number of neighbors of each field--i.e. fields that one can
    // move to from current field.
    abstract public byte getCountNeighborIndices();

    // mandatory directions
    abstract public byte getDirUp();
    abstract public byte getDirDown();
    abstract public byte getDirLeft();
    abstract public byte getDirRight();

    abstract public byte getReciprocal(byte dir1);

    // return an index, or -1 if there's no neighbor in that direction
    abstract public byte getNeighbor(byte fieldIdx, byte dir1);

    // where should the cursor be placed when the game starts
    abstract public byte getInitialPos();

    public byte getNumberOfFields() {
	return (byte)fields.length;
    }

    // Useful for MineField subclasses that have a notion of width and
    // height.
    protected int boundWrap(int c, int s) {
	if (c < 0)
	    return s + c % s;
	else if (c >= s)
	    return c % s;
	else
	    return c;
    }

    protected int bound(boolean wrap, int c, int s) {
	if (wrap)
	    return boundWrap(c, s);
	else {
	    if (c < 0)
		return 0;
	    else if (c >= s)
		return s - 1;
	    else
		return c;
	}
    }


    // - - - - - - - - - - - - - - - - -
    // Field handling
    // - - - - - - - - - - - - - - - - -

    public Field getField(byte fieldIdx) {
	return fields[fieldIdx];
    }

    public Field getNeighbor(Field f, byte dir1) {
	byte fieldIdx2 = getNeighbor(f.fieldIdx, dir1);
	if (fieldIdx2 < 0)
	    return null;
	else
	    return getField(fieldIdx2);
    }

    public void allFields(FieldEnumerator e) {
	for (int i = 0; i < fields.length; ++i)
	    e.field(fields[i]);
    }

    public void allNeighbors(byte fieldIdx, FieldEnumerator e) {
	byte count = getCountNeighborIndices();
	for (byte nIdx = 0; nIdx < count; ++nIdx) {
	    byte fieldIdx2 = getNeighbor(fieldIdx, nIdx);
	    if (fieldIdx2 < 0)
		continue;
	    if (!e.field(getField(fieldIdx2)))
		break;
	}
    }

    public boolean uncover(byte idx1)
	throws IllegalArgumentException
    {
	switch (fields[idx1].uncover()) {
	case UNCOVER_BUSTED:
	    busted = true;
	    return true;
	case UNCOVER_NEW:
	    checkWin();
	    return true;
	}
	return false;
    }

    // - - - - - - - - - - - - - - - - -
    // Participants
    // - - - - - - - - - - - - - - - - -

    public ParticipantSelector getParticipantSelector() {
	return parSel;
    }

    // Answers whether fieldIdx1 is participant of fieldIdx2
    public boolean isParticipantOf(final byte fieldIdx1, byte fieldIdx2) {

	class E implements FieldEnumerator {
	    public boolean found = false;
	    public boolean field(Field f) {
		if (f.fieldIdx == fieldIdx1)
		    found = true;
		return !found;
	    }
	}
	E e = new E ();
	getField(fieldIdx2).allParticipants(e);
	return e.found;
    }

    // - - - - - - - - - - - - - - - - -
    // Game state
    // - - - - - - - - - - - - - - - - -

    public boolean isBusted() {
	return busted;
    }

    public boolean hasWon() {
	return won;
    }

    public boolean gameOver() {
	return hasWon() || isBusted();
    }

    private void checkWin() {
	if (busted)
	    return;
	for (byte idx1 = 0; idx1 < fields.length; ++idx1) {
	    boolean mine = fields[idx1].hasMine();
	    boolean uncovered = uncovered(fields[idx1]);
	    if (mine && uncovered)
		return;
	    if (!mine && !uncovered)
		return;
	}

	allFields(new FieldEnumerator() {
		public boolean field(Field f) {
		    if (hasMine(f) && !flagged(f))
			toggleFlag(f);
		    return true;
		}
	    });

	won = true;
    }

    // - - - - - - - - - - - - - - - - -
    // Encoding
    // - - - - - - - - - - - - - - - - -

    private boolean getEncodingBit(byte idx1) {
	return idx1 < fields.length ? fields[idx1].hasMine() : false;
    }
    public String getEncoding() {
	// Will fail loudly if fields.length % 5 != 0
	StringBuffer buf = new StringBuffer();
	for (byte idx1 = 0; idx1 < fields.length; ) {
	    boolean a1 = fields[idx1++].hasMine();
	    boolean a2 = getEncodingBit(idx1++);
	    boolean a3 = getEncodingBit(idx1++);
	    boolean a4 = getEncodingBit(idx1++);
	    boolean a5 = getEncodingBit(idx1++);
	    int code = (a1 ? 16 : 0) + (a2 ? 8 : 0)
		+ (a3 ? 4 : 0) + (a4 ? 2 : 0) + (a5 ? 1 : 0);
	    char c = "abcdefghijklmnop".charAt(code % 16);
	    if (code > 15)
		c = Character.toUpperCase(c);
	    buf.append(c);
	}
	buf.append(';');
	return buf.toString();
    }

    public int hashCode() {
	int hash = 0;
	for (byte idx1 = 0; idx1 < fields.length; ++idx1) {
	    int h0 = ((hash >>> 24) & 0xff);
	    hash = hash << 1;
	    hash = hash ^ h0;
	    if (fields[idx1].hasMine())
		hash |= 1;
	    else
		hash &= ~1;
	}
	return hash;
    }


    // - - - - - - - - - - - - - - - - -
    // Painting
    // - - - - - - - - - - - - - - - - -

    public void paint(final MineSweeperDisplay d) {
	d.beforeDrawing();
	allFields(new FieldEnumerator() {
		public boolean field(Field f) {
		    paintField(f, d);
		    return true;
		}
	    });
	d.afterDrawing();
    }

    public abstract void paintField(Field f, MineSweeperDisplay d);

    public void paint(boolean full) {
	paint(new StringGridDisplay(System.out, full));
    }

    public void paint() {
	paint(true);
    }

    // Answer maximum coordinate that will be used when calling
    // MineSweeperDisplay methods.
    abstract public int getPaintWidthRequest();
    abstract public int getPaintHeightRequest();
}
