/*
 * Copyright 2008,2009,2010,2011,2012 Petr Machata
 *
 * This file is part of MicroSweeper.
 *
 * MicroSweeper is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * MicroSweeper is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MicroSweeper.  If not, see <http://www.gnu.org/licenses/>.
 */

import javax.microedition.lcdui.*;
import javax.microedition.midlet.*;
import java.lang.Math;

public class MineSweeperCanvas
    extends Canvas
    implements CommandListener
{
    private final MicroSweeper parent;
    private boolean building = false;

    private final MineField mineField;
    private final MineFieldBuilder builder;

    private final Command backCommand;
    private final Command flagCommand;
    private final Command restartCommand;

    // Start at the corner, where there is the best chance to get a
    // cascade on the first click.
    private byte cursor;
    private boolean showParticipantHints = true;
    private boolean fullPlan;

    private class FontBestMatch {
	private int diff = 0;
	private Font bestMatch = null;
	final private int wantHeight;

	private void checkBestMatch(int face, int style, int size) {
	    Font font = Font.getFont(face, style, size);
	    int baseline = font.getBaselinePosition();
	    int d = baseline - wantHeight;

	    if (d > 0)
		d *= 5; // Penalize too large a font, but still select
			// it if tiny font is the only alternative.

	    if (bestMatch == null || Math.abs(d) < Math.abs(diff))
		bestMatch = font;
	}

	public FontBestMatch(int wantHeight) {
	    this.wantHeight = wantHeight;
	}

	public FontBestMatch checkFont(int face, int style) {
	    checkBestMatch(face, style, Font.SIZE_SMALL);
	    checkBestMatch(face, style, Font.SIZE_MEDIUM);
	    checkBestMatch(face, style, Font.SIZE_LARGE);
	    return this;
	}

	public Font getBestMatch() {
	    return bestMatch;
	}
    }

    public MineSweeperCanvas(MicroSweeper parent, MineFieldBuilder builder,
			     boolean fullPlan) {
	this.parent = parent;
	this.builder = builder;
	this.mineField = builder.mineField;
	this.cursor = mineField.getInitialPos();
	this.fullPlan = fullPlan;

	if (fullPlan)
	    startBuilding(null);

	addCommand(backCommand = new Command("Back", Command.BACK, 0));
	addCommand(flagCommand = new Command("Flag", Command.ITEM, 0));
	restartCommand = new Command("Restart", Command.ITEM, 1);
	setCommandListener(this);
    }

    private abstract class CanvasBase {
	/// Conceptually, we allow painting to a canvas of a
	/// dimensions requested by a board.  These are pixels per
	/// unit requested.
	protected final int r;

	public CanvasBase() {
	    final MineField mf = MineSweeperCanvas.this.mineField;
	    int resX = getWidth() / mf.getPaintWidthRequest();
	    int resY = getHeight() / mf.getPaintHeightRequest();
	    this.r = resX < resY ? resX : resY;
	}

	public void drawField(Field f, int fx, int fy, int fr) {
	    int x0 = this.r * fx;
	    int y0 = this.r * fy;
	    int w0 = this.r * (fx + fr) - x0;
	    int h0 = this.r * (fy + fr) - y0;

	    doDrawField(f, x0, y0, w0, h0);
	}

	abstract protected void doDrawField(Field f,
					    int x0, int y0, int w0, int h0);
    }

    private class CanvasDisplay
	extends CanvasBase
	implements MineSweeperDisplay
    {
	protected final Graphics g;
	protected final boolean isColor;

	/// Mine count font cache and the height that it was selected
	/// for.
	protected Font mineCountFont = null;
	protected int mineCountH = 0;

	public CanvasDisplay(Graphics g) {
	    this.g = g;

	    // See if we have colors.
	    Display disp = Display.getDisplay(parent);
	    this.isColor = disp.isColor();
	}

	protected boolean shouldDrawBuilderProgress(byte idx1, int progress) {
	    return progress > idx1;
	}

	private void paintFlag(int x0, int y0, int x1, int y1) {
	    int w0 = x1 - x0 + 1;
	    int h0 = y1 - y0 + 1;

	    g.setColor(0);
	    int xm = (x0 + x1) / 2;
	    g.drawLine(xm - 2, y0 + 5, xm - 2, y1 - 5);
	    g.drawLine(xm - 1, y0 + 5, xm - 1, y1 - 5);
	    g.drawLine(xm, y0 + 5, xm, y1 - 5);

	    if (Display.getDisplay(parent).isColor())
		g.setColor(0xff0000);
	    else
		g.setColor(0xffffff);
	    g.fillRect(xm, y0 + 5, w0 / 3, h0 / 4);
	}

	private void paintCircle(int x0, int y0, int x1, int y1) {
	    int w0 = x1 - x0;
	    int h0 = y1 - y0;

	    g.setColor(0);
	    for (int i = 2; i < 5; ++i)
		g.drawArc(x0+i, y0+i, w0 - 2*i, h0 - 2*i, 0, 360);
	}

	private void paintMineMark(int x0, int y0, int x1, int y1) {
	    if (fullPlan)
		paintCircle(x0, y0, x1, y1);
	    else
		paintFlag(x0, y0, x1, y1);
	}

	private void paintCross(int x0, int y0, int x1, int y1) {

	    if (Display.getDisplay(parent).isColor())
		g.setColor(0xff0000);
	    else
		g.setColor(0);

	    g.drawLine(x0+2, y0+2, x1-2, y1-2);
	    g.drawLine(x0+3, y0+2, x1-2, y1-3);
	    g.drawLine(x0+2, y0+3, x1-3, y1-2);

	    g.drawLine(x0+2, y1-2, x1-2, y0+2);
	    g.drawLine(x0+3, y1-2, x1-2, y0+3);
	    g.drawLine(x0+2, y1-3, x1-3, y0+2);
	}

	private void paintNumber(Field f, int x0, int y0, int w0, int h0) {
	    String s = f.render();
	    if (s == "")
		return;

	    if (isColor)
		g.setColor(f.getColor());
	    else
		g.setColor(0);

	    int center = x0 + w0/2;
	    int d = (h0 - mineCountFont.getBaselinePosition()) / 2;
	    g.drawString(s, center, y0 + d,
			 Graphics.HCENTER | Graphics.TOP);
	}

	private void paintArrow(int x1, int y1, int x2, int y2,
				int x3, int y3, boolean vertical) {
	    int x4 = x1;
	    int y4 = y2;

	    int x5 = vertical ? x2 : x1;
	    int y5 = vertical ? y2 : y1;

	    if (x1 > x2) {
		int tmp = x1;
		x1 = x2;
		x2 = tmp;
	    }

	    if (y1 > y2) {
		int tmp = y1;
		y1 = y2;
		y2 = tmp;
	    }

	    g.setColor(0x00808080);
	    g.fillRect(x1, y1, x2 - x1, y2 - y1);

	    g.setColor(0);
	    g.drawRect(x1, y1, x2 - x1, y2 - y1);

	    g.setColor(0x00808080);
	    g.fillTriangle(x4, y4, x5, y5, x3, y3);

	    g.setColor(0);
	    g.drawLine(x5, y5, x3, y3);
	    g.drawLine(x3, y3, x4, y4);
	}

	private void paintIcon(Field f, int x0, int y0, int w0, int h0) {
	    switch (f.getIcon()) {
	    case Field.iconNone:
		return;

	    case Field.iconCircle:
		g.setColor(0xb0b0b0);
		g.fillArc(x0 + 2, y0 + 2, w0 - 4, h0 - 4, 0, 360);
		g.setColor(0);
		g.drawArc(x0 + 2, y0 + 2, w0 - 4, h0 - 4, 0, 360);
		return;

	    case Field.iconArrowDown:
		paintArrow(x0 + 2, y0 + 2, x0 + w0 - 4, y0 + 2*h0/3,
			   x0 + w0/2, y0 + h0 - 2, true);
		return;

	    case Field.iconArrowRight:
		paintArrow(x0 + 2*w0/3, y0 + h0 - 4, x0 + 2, y0 + 2,
			   x0 + w0 - 2, y0 + h0/2, false);
		return;
	    }
	}

	private void paintConsistency(Field f, int x0, int y0, int w0, int h0) {
	    if (!f.consistent()) {
		if (isColor)
		    g.setColor(0xff0000);
		else
		    g.setColor(0);
		g.drawRect(x0 + 1, y0 + 1, w0 - 2, h0 - 2);
	    }
	}

	protected void doDrawField(Field f, int x0, int y0, int w0, int h0) {
	    int x1 = x0 + w0 - 1;
	    int y1 = y0 + h0 - 1;

	    if (mineCountFont == null || h0 != mineCountH) {
		mineCountFont = new FontBestMatch(h0)
		    .checkFont(Font.FACE_MONOSPACE, Font.STYLE_BOLD)
		    .getBestMatch();
		mineCountH = h0;
	    }
	    g.setFont(mineCountFont);

	    if (!f.uncovered() || building) {

		g.setColor(0x00808080);
		g.fillRect(x0, y0, w0, h0);

		// Shade...
		g.setColor(0x00505050);
		g.drawLine(x0+1, y1, x1, y1);
		g.drawLine(x1, y1, x1, y0+1);

		g.setColor(0x00707070);
		g.drawLine(x0+2, y1-1, x1-1, y1-1);
		g.drawLine(x1-1, y1-1, x1-1, y0+2);

		// Light...
		g.setColor(0x00b0b0b0);
		g.drawLine(x0, y1-1, x0, y0);
		g.drawLine(x0, y0, x1-1, y0);

		g.setColor(0x00909090);
		g.drawLine(x0+1, y1-2, x0+1, y0+1);
		g.drawLine(x0+1, y0+1, x1-2, y0+1);

		if (building
		    && shouldDrawBuilderProgress
			(f.fieldIdx, builder.getProgressMines()))
		    paintMineMark(x0, y0, x1, y1);
		else if (f.flagged()) {
		    if (mineField.gameOver() && !f.hasMine())
			paintCross(x0, y0, x1, y1);
		    else
			paintMineMark(x0, y0, x1, y1);
		}
		if (fullPlan && !building) {
		    paintNumber(f, x0, y0, w0, h0);
		    paintConsistency(f, x0, y0, w0, h0);
		}

	    } else {
		g.setColor(0x00909090);
		g.fillRect(x0, y0, w0, h0);

		g.setColor(0x00a0a0a0);
		g.drawRect(x0, y0, w0, h0);

		paintIcon(f, x0, y0, w0, h0);

		if (!f.hasMine() && !mineField.hasWon()) {
		    paintNumber(f, x0, y0, w0, h0);
		    paintConsistency(f, x0, y0, w0, h0);
		}
	    }

	    if (mineField.isBusted() && !f.flagged() && f.hasMine()) {
		g.setColor(0);
		g.fillArc(x0 + 5, y0 + 5, w0 - 11, h0 - 11, 0, 360);
	    }

	    // Draw cursor and participant hints.
	    if (!mineField.gameOver()) {
		if (f.fieldIdx == cursor) {
		    g.setColor(0x00d0d0d0);
		    g.drawRect(x0 + 2, y0 + 2, w0 - 5, h0 - 5);
		}
		else if (showParticipantHints
			 && !f.uncovered()
			 && mineField.isParticipantOf(f.fieldIdx, cursor)) {
		    g.setColor(0x00ffffff);
		    g.drawRect(x0 + w0 / 2 - 1, y0 + h0 / 2 -1, 3, 3);
		}
	    }
	}

	public void beforeDrawing() {
	    g.setColor(0);
	    g.fillRect(0, 0, getWidth(), getHeight());
	}

	public void afterDrawing() {
	}
    }

    class RepaintDisplay
	extends CanvasBase
	implements MineSweeperDisplay
    {
	protected void doDrawField(Field f, int x0, int y0, int w0, int h0) {
	    // repaint more to accomodate for corners
	    repaint(x0 - 3, y0 - 3, w0 + 6, h0 + 6);
	}

	public void beforeDrawing() {}
	public void afterDrawing() {}
    }

    protected void paint(Graphics g) {
	// The display has to be reconstructed on each repaint,
	// because in general we don't know if graphics size didn't
	// change (plus the old reference to `g' would probably be
	// invalid).
	mineField.paint(new CanvasDisplay(g));
    }

    private void repaintField(Field f) {
	mineField.paintField(f, new RepaintDisplay());
    }

    private boolean moveCursor(byte dir1) {
	byte nc = mineField.getNeighbor(cursor, dir1);
	if (nc < 0 || nc == cursor)
	    return false;
	moveCursorTo(nc);
	return true;
    }

    private void moveCursorTo(byte nc) {
	FieldEnumerator participantRepainter
	    = new FieldEnumerator() {
		    public boolean field(Field f) {
			repaintField(f);
			return true;
		    }
		};

	Field cursorField = mineField.getField(cursor);
	repaintField(cursorField);
	cursorField.allParticipants(participantRepainter);

	Field ncField = mineField.getField(nc);
	repaintField(ncField);
	ncField.allParticipants(participantRepainter);

	cursor = nc;
    }

    // Repaint all fields that contain F as participant.
    private void repaintObservers(final Field f) {
	mineField.allFields(new FieldEnumerator() {
		public boolean field(Field g) {
		    class NeedsRepaint implements FieldEnumerator {
			boolean needs = false;
			public boolean field(Field h) {
			    if (f == h) {
				needs = true;
				return false;
			    }
			    return true;
			}
		    }
		    NeedsRepaint needsRepaint = new NeedsRepaint();
		    g.allParticipants(needsRepaint);
		    if (needsRepaint.needs)
			repaintField(g);
		    return true;
		}
	    });
    }

    private void toggleFlag(byte idx1) {
	Field f = mineField.getField(idx1);
	f.toggleFlag();
	repaintField(f);
	repaintObservers(f);
    }

    private boolean uncover(final byte idx1) {
	Field f = mineField.getField(idx1);
	if (!f.uncovered() && !f.flagged()) {
	    mineField.uncover(idx1);

	    if (f.countMines() == 0 && !f.hasMine())
		f.allParticipants(new FieldEnumerator() {
			public boolean field(Field f) {
			    if (!f.uncovered())
				uncover(f.fieldIdx);
			    return true;
			}
		    });

	    repaintField(f);
	    repaintObservers(f);
	    return true;

	} else
	    return false;
    }

    private void doneBuilding(byte idx1) {
	building = false;
	repaint();
	System.out.println("done building");

	/* Open all trivial constraints.  */
	mineField.allFields(new FieldEnumerator() {
		public boolean field(Field f) {
		    if (f.countMines() == 0
			&& (f.uncovered() || fullPlan))
			f.allParticipants(new FieldEnumerator() {
				public boolean field(Field f) {
				    uncover(f.fieldIdx);
				    return true;
				}
			    });
		    return true;
		}
	    });

	if (!fullPlan)
	    fire(idx1);
    }

    private void fire(byte idx1) {
	uncover(idx1);
	if (mineField.gameOver()) {
	    repaint();
	    removeCommand(flagCommand);
	    addCommand(restartCommand);
	}
    }

    private void startBuilding(final Field f0) {
	building = true;
	showParticipantHints = false;

	class RepaintThread extends Thread {
	    private boolean done = false;
	    public void doneBuilding() {
		done = true;
	    }
	    public void run() {
		int last = 0;
		while (!done) {
		    try {
			sleep(500);
		    } catch (InterruptedException e) {
			// never mind...
		    }
		    int prog = Math.max(builder.getProgressWalls(),
					builder.getProgressMines());
		    int max = Math.max(prog, last);
		    int min = Math.min(prog, last);
		    for (int i = min; i <= max; ++i)
			repaintField(mineField.getField((byte)i));
		    last = prog;
		}
	    }
	}

	class BuilderThread extends Thread {
	    RepaintThread rpt;
	    public BuilderThread(RepaintThread t) {
		rpt = t;
	    }
	    public void run() {
		builder.initialize(f0);
		rpt.doneBuilding();
		MineSweeperCanvas.this.doneBuilding(cursor);
	    }
	}

	RepaintThread rpt = new RepaintThread();
	BuilderThread bt = new BuilderThread(rpt);
	rpt.start();
	bt.start();
    }

    private void fireAction() {
	if (!builder.isInitialized())
	    startBuilding(mineField.getField(cursor));
	else
	    fire(cursor);
    }

    private void flagAction() {
	if (!mineField.getField(cursor).uncovered()
	    && builder.isInitialized())
		toggleFlag(cursor);
    }

    protected void keyPressed(int keyCode) {
	if (mineField.gameOver() || building)
	    return;

	switch (getGameAction(keyCode)) {
	case UP:
	    moveCursor(mineField.getDirUp());
	    break;

	case DOWN:
	    moveCursor(mineField.getDirDown());
	    break;

	case LEFT:
	    moveCursor(mineField.getDirLeft());
	    break;

	case RIGHT:
	    moveCursor(mineField.getDirRight());
	    break;

	case FIRE:
	    fireAction();
	    break;
	}
    }

    class FieldLocator
	extends CanvasBase
	implements MineSweeperDisplay
    {
	private final int pointX, pointY;
	public Field field = null;
	public FieldLocator(int x, int y) {
	    pointX = x;
	    pointY = y;
	}

	protected void doDrawField(Field f, int x0, int y0, int w0, int h0) {
	    if (pointX >= x0 && pointX < x0 + w0
		&& pointY >= y0 && pointY < y0 + h0)
		this.field = f;
	}

	public void beforeDrawing() {}
	public void afterDrawing() {}
    }

    class PressThread extends Thread {
	private final long pressTime = System.currentTimeMillis();
	public boolean done = false;
	public boolean fire;

	public PressThread(boolean fire) {
	    this.fire = fire;
	}

	public long passedTimeMillis() {
	    return System.currentTimeMillis() - pressTime;
	}

	public void run() {
	    while (!done) {
		long rem = 750 - passedTimeMillis();

		if (rem <= 0) {
		    done = true;
		    flagAction();
		    break;
		}

		try {
		    sleep(rem);
		} catch (InterruptedException e) {
		    // never mind...
		}
	    }
	}
    }
    public PressThread pressThread = null;

    protected void pointerPressed(int x, int y) {
	if (mineField.gameOver() || building)
	    return;

	FieldLocator locator = new FieldLocator(x, y);
	mineField.paint(locator);
	if (locator.field != null) {
	    byte idx = locator.field.fieldIdx;
	    pressThread = new PressThread(cursor == idx);
	    moveCursorTo(idx);
	    pressThread.start();
	}
    }

    protected void pointerReleased(int x, int y) {
	if (pressThread == null)
	    return;

	if (pressThread.fire && pressThread.passedTimeMillis() < 250) {
	    // If the user "dragged away" from the source field, then
	    // don't open it.  He's probably having second thoughts.
	    FieldLocator locator = new FieldLocator(x, y);
	    mineField.paint(locator);
	    if (locator.field != null && locator.field.fieldIdx == cursor)
		fireAction();
	}

	pressThread.done = true;
	pressThread = null;
    }

    public void commandAction(Command c, Displayable s) {
	if (c == backCommand) {
	    if (mineField.gameOver())
		parent.exitGame();
	    else
		parent.showMenu(this);
	} else if (c == restartCommand) {
	    parent.restartGame();
	} else if (c == flagCommand && !mineField.gameOver()) {
	    flagAction();
	}
    }

    public void stop() {
	builder.stop();
    }
}
